from sklearn.feature_extraction.text import TfidfVectorizer, CountVectorizer
from sklearn.feature_selection import SelectKBest, chi2
from sklearn import metrics, ensemble, linear_model, svm
from numpy import log, ones, array, zeros, mean, std, repeat
import numpy as np
import scipy.sparse as sp
import re
import csv
from time import time

# All the files should be in the directory you are running the code in.
# DIR_PATH should be set to the path of the directory containing all the files.
# If you want to test on a new file , the TEST_FILE should be set to the same.
# the function run() is the main function which saves the predictions in the PREDICTION_FILE, each prediction being a number betweeen 0 and 1.
# The binary predictions {0,1} are saved in BINARY_PREDICTION_FILE
# Code run on Linux 32-bit with python 2.7.

# there are three files which can be used to train the data, namely : train.csv, test_with_solutions.csv and test_file.csv

DIR_PATH = "/home/chetan/AI/code/codes/"

#TRAIN_FILE      = DIR_PATH + "train.csv"
TRAIN_FILE      = DIR_PATH + "Train1.csv"
TEST_SOL_FILE   = DIR_PATH + "Train2.csv" 	 # This is also used for training, together with TRAIN_FILE
BADWORDS_FILE   = DIR_PATH + "badwords.txt"                      # The mapping to be used in normalizing

#TEST_FILE       = DIR_PATH + "test_file.csv"        		 # set this to the new test file name
TEST_FILE       = DIR_PATH + "Test.csv"
PREDICTION_FILE = DIR_PATH + "predictions_new.csv"                   # Predictions will be written here 
BINARY_PREDICTION_FILE = DIR_PATH + "binary_new.csv"     # Predictions more than 0.5 would be written as 1 , rest as 0

def normalize(f):
    f = [x.replace('\xe0\xa4\x95\xe0\xa4\xbe\xe0\xa4\xb2\xe0\xa5\x80\xe0\xa4\xa8', '\xe0\xa4\x95\xe0\xa4\xbe\xe0\xa4\xb2\xe0\xa5\x80\xe0\xa4\xa8') for x in f]
    f = [x.replace("\n"," ") for x in f]
    
    f = [x.replace('\\',' ') for x in f]
    f = [x.replace("\n"," ") for x in f]
    f = [x.replace("\\t"," ") for x in f]        
    f = [x.replace("\\xa0"," ") for x in f]
    f = [x.replace("\\xc2"," ") for x in f]
 
    return f


def cleaning(data, labels, ntrain, true_or_false = False):
    f = data;
    if true_or_false:
        f = normalize(f)
    return f
		
	
########
# extracts features using count of words, characters and n-grams 
# extracts features using count of words, characters and n-grams 
def ngrams(data, labels, ntrain, mn=1, mx=1, nm=100, binary = False, stopwords = False, analyzer_char = False):
    f = data
    f = normalize(f)
    
    ftrain = f[:ntrain]
    ftest  = f[ntrain:]
    y_train = labels[:ntrain]
    
    t0 = time()
    analyzer_type = 'word'
    if analyzer_char:
        analyzer_type = 'char'
        

    vectorizer = TfidfVectorizer({mn,mx},sublinear_tf=True,analyzer=analyzer_type)

    print "extracting ngrams for [%d,%d]" % (mn,mx)
    
    X_train = vectorizer.fit_transform(ftrain)
    X_test = vectorizer.transform(ftest)
    
    print "done in %fs" % (time() - t0), X_train.shape, X_test.shape

    y = array(y_train)    
    
    if X_train.shape[1]>200:
        t0 = time()
        print "chi-Squared test"
        ch2 = SelectKBest(chi2, k = 200)
        X_train = ch2.fit_transform(X_train, y)
        X_test = ch2.transform(X_test)
        assert sp.issparse(X_train)        

    print "Extracting best features by a chi-squared test.. ", X_train.shape, X_test.shape    
    return X_train, y, X_test


def second_person(data, labels, ntrain):
    print "Second Person"
    g = [x.lower().replace("\xe0\xa4\x86\xe0\xa4\xaa"," ABCD ").replace("\xe0\xa4\xa4\xe0\xa5\x81\xe0\xa4\xae"," ABCD ").split("ABCD")[1:] for x in data]

    f = []
    for x in g:
        fts = " "
        x = normalize(x)
        for y in x:
            w = y.strip().replace("?",".").split(".")
            fts = fts + " " + w[0]        
        f.append(fts)
    
    X_trn, y_trn, X_tst = ngrams(f, labels, ntrain, 1, 1, 100)
    return X_trn, y_trn, X_tst  

def negation_feature(data, labels, ntrain):
    print "Negation Feature"
    g = [x.lower().replace("\xe0\xa4\xa8"," XYZ ").replace("\xe0\xa4\xa8\xe0\xa4\xb9\xe0\xa5\x80\xe0\xa4\x82"," XYZ ").split("XYZ")[1:] for x in data]

    f = []
    for x in g:
        fts = " "
        x = normalize(x)
        for y in x:
            w = y.strip().replace("?",".").split(".")
            fts = fts + " " + w[0]        
        f.append(fts)
    
    X_trn, y_trn, X_tst = ngrams(f, labels, ntrain, 1, 1, 100)
    return X_trn, y_trn, X_tst 

# the main function 
def run():
    t0 = time()

    train_data = readCsv(TRAIN_FILE)
    train2_data = readCsv(TEST_SOL_FILE)
    #train_data = train_data + train2_data                       # the two files are combined and used as training set
    
    labels  = array([int(x[1]) for x in train_data])    	#these are the labels in the training set
        
    train  = [x[2] for x in train_data]				#training data comments in forum

    test_data = readCsv(TEST_FILE)
    test_data = [x[2] for x in test_data]
#data = train	    
    data = train + test_data
    n = len(data)
    ntrain = len(train)

    # The number of ngrams to select was optimized by Cross Validation
    X_train1, y_train, X_test1 = ngrams(data, labels, ntrain, 1, 1, 2000)
    X_train2, y_train, X_test2 = ngrams(data, labels, ntrain, 2, 2, 4000)
    X_train3, y_train, X_test3 = ngrams(data, labels, ntrain, 3, 3, 100)    
    X_train4, y_train, X_test4 = ngrams(data, labels, ntrain, 4, 4, 1000, analyzer_char = True)    
    X_train5, y_train, X_test5 = ngrams(data, labels, ntrain, 5, 5, 1000, analyzer_char = True)    
    X_train6, y_train, X_test6 = ngrams(data, labels, ntrain, 3, 3, 2000, analyzer_char = True)    

    X_train7, y_train, X_test7 = second_person(data, labels, ntrain)
    X_train8, y_train, X_test8 = negation_feature(data, labels, ntrain)

    #stack all fatures together increasing the size of feature vector
    X_tn = sp.hstack([X_train1, X_train2, X_train3, X_train4, X_train5, X_train6, X_train7])
    X_tt = sp.hstack([X_test1,  X_test2,  X_test3, X_test4, X_test5, X_test6, X_test7])
    
    print "######## Total time for feature extraction: %fs" % (time() - t0), X_tn.shape, X_tt.shape
    
    predictions = runClassifiers(X_tn, labels, X_tt)
    
    write_submission(predictions, PREDICTION_FILE)
    write_binary_submission(predictions, BINARY_PREDICTION_FILE)
        
    print "Predictions written to:", PREDICTION_FILE
    print "Binary Predictions written to:", BINARY_PREDICTION_FILE

   
    # the percentage of correct predictions
    percentage_accuracy = accuracy()
    print "Accuracy = %.10f" % (percentage_accuracy)
    
    #precision and recall 
    precision, recall = precision_and_recall()
    #precision = norm(precision) 
    print "Precision = %.10f, Recall = %.10f" % (precision,recall)
    
def runClassifiers(X_train, y_train, X_test, y_test = None):
    
    models = [  linear_model.LogisticRegression(C=3), \
                svm.SVC(C=0.3,kernel='linear',probability=True) ]
                
    dense = [False, False, True, True]    # if the model needs dense matrix
    
    X_train_dense = X_train.todense()
    X_test_dense  = X_test.todense()
    
    preds = []
    for ndx, model in enumerate(models):
        t0 = time()
        print "Training: ", model, 20 * '_'        
        if dense[ndx]:
            model.fit(X_train_dense, y_train)
            pred = model.predict_proba(X_test_dense)    
        else:
            model.fit(X_train, y_train)
            pred = model.predict_proba(X_test)
        print "Training time: %0.3fs" % (time() - t0)
        preds.append(array(pred[:,1]))
        
      
    final_pred = preds[0]*0.4+preds[1]*0.6
    
    return final_pred


def readCsv(fname, skipFirst=True, delimiter = ","):
    reader = csv.reader(open(fname,"rb"),delimiter=delimiter)
    
    rows = []
    count = 1
    for row in reader:
        if not skipFirst or count > 1:      
            rows.append(row)
        count += 1
    return rows

def write_submission(x,filename):
    wtr = open(filename,"w")
    for i in range(len(x)):
        wtr.write(format(x[i],"0.10f"))
        wtr.write("\n")
    wtr.close()

def write_binary_submission(x,filename):
    wtr=open(filename,"w")
    for i in range(len(x)):
        if (x[i] >= 0.5):
	    wtr.write("1")
            wtr.write("\n")
	else:
            wtr.write("0")
            wtr.write("\n")
    wtr.close()
   

def accuracy():
    to_check_file = readCsv(BINARY_PREDICTION_FILE)
    to_check_data = [x[0] for x in to_check_file]
    
    print len(to_check_data)
    
    test_data = readCsv(TEST_FILE)
    answers = [x[1] for x in test_data]
    print len(answers)        
    count = 0
    for i in range(len(answers)-1):
        if int(answers[i+1]) == int(to_check_data[i]) :   # matching the entries in the prediction file and the known results
            count = count + 1 
	    #print count    
    return (float(count)/(len(answers)-1))*100


def precision_and_recall():
    predicted_file = readCsv(BINARY_PREDICTION_FILE)
    predicted_data = [x[0] for x in predicted_file]

    observed_file = readCsv(TEST_FILE)
    observed_data = [x[1] for x in observed_file]

    true_positives=0
    false_positives=0
    true_negatives=0
    false_negatives=0

    for i in range(len(observed_data)-1):
        if (int(observed_data[i+1])==0):
	    if(int(predicted_data[i])==0):
                true_negatives+=1
            else:
                false_positives+=1
        else:
	    if(int(predicted_data[i])==1):
                true_positives+=1
            else:
                false_negatives+=1
    precision=float(true_positives)/(true_positives+false_positives)
    recall= float(true_positives)/(true_positives+false_negatives)
    #print "Accuracy using precision and recall =" , float(true_positives+true_negatives)/(true_positives+false_positives+true_negatives+false_negatives)
    return precision,recall
        
if __name__=="__main__":
    run()
